%----------------------------------------------------------------------------
% BBQ method for box-constrained optimization
% 
%               min  f(x)  s.t. LB <= x <= UB
% 
% You should provide two functions: fobj for function value, and grad for gradient   
% 
% Stop condition: max(abs(pg - x))<=tol
% 
% Using as 
%           [out, hist] = BBQ_box(x0, fobj, grad, LB, UB)
%           [out, hist] = BBQ_box(x0, fobj, grad, LB, UB, opts, LinesearchParms)
% 
% Reference: 
%   [1] Yakui Huang, Yu-Hong Dai, Xin-Wei Liu, Equipping Barzilai-Borwein method 
%       with two dimensional quadratic termination property, 2021.      
%
% Authors:
%       Yakui Huang, Yu-Hong Dai, Xin-Wei Liu
%       contact hyk@hebut.edu.cn; dyh@lsec.cc.ac.cn; math-lxw@hebut.edu.cn
% 
% Version: 1.0
% Last Update: Jan 8, 2021
%----------------------------------------------------------------------------

function [out, hist] = BBQ_box(x0, fobj, grad, LB, UB, varargin)

%----------------------------------------------------------------------------
% Input:
%          x0   --- initial guess
%         fobj  --- objective function: f = fobj(x)
%         grad  --- gradient function: g = grad(x)
%          LB   --- lower bound of variables
%          UB   --- upper bound of variables
%         opts  --- structure for containing input parameters (optional) with fields:
%            PrintLevel ---   0. no print out
%                             1. print final values
%                             2. print values for every iterations
%               MaxIter --- max number of iterations
%               MaxFunc --- max number of function evaluations
%                   tol --- stop control for the projected gradient
%                 bbmin --- lower bound of bb-like stepsizes
%                 bbmax --- upper bound of bb-like stepsizes                              
%                   tau --- adaptive parameter
%                  gama --- for updating adaptive parameter
%                record --- whether record history information: 0 (N) or 1 (Y)
%      LinesearchParms  --- structure for containing line search parameters (optional) with fields:
%                 delta --- parameter for function value reduction
%                 sigma --- parameter for reducing steplength
%                     M --- number of record function values
%                maxlsf --- max number of function evaluations in line search          
% 
%----------------------------------------------------------------------------  
% Output:
%         out --- output information with fields:
%               out.code --- termination code: 0. solution found
%                                              1. maximum number of function evaluations exceeded
%                                              2. maximum number of iterations exceeded
%                                              3. line search fails
%               out.iter --- number of iterations
%              out.nfunc --- number of function evaluations
%              out.ngrad --- number of gradient evaluations
%                  out.x --- solution
%                out.obj --- function value at solution
%             out.pgnorm --- gradient norm at solution
%            out.cputime --- cpu time
%        hist --- history information with fields:
%               hist.bbs --- history of bb-like stepsizes 
%              hist.step --- history of steplengths obtained by line search
%               hist.obj --- history of function values
%            hist.pgnorm --- history of gradient norms
% -----------------------------------------------------------------------

%% ------------------------------------------------------------------------
% check input parameters
if (nargin < 5)
    error('at least three inputs: [out, hist] = BBQ_box(x0, fobj, grad, LB, UB)');
elseif (nargin < 6)
    opts = [];
    LinesearchParms = [];
elseif (nargin < 7)
    opts = varargin{1};
    LinesearchParms = [];
elseif (nargin == 8)
    opts = varargin{1};
    LinesearchParms = varargin{2};
end

if ( isempty(LB) )
    LB = -inf.*ones(size(x0));
end
if ( isempty(UB) )
    UB = inf.*ones(size(x0));
end

%% -------------------------------------------------------------------------
% initialize parameters for options
% print level
if isfield(opts,'PrintLevel')
    if ( opts.PrintLevel < 0 )
        error('opts.PrintLevel should be nenagtive integer \n');
    end
    if ( opts.PrintLevel ~= fix(opts.PrintLevel) )
        error('opts.PrintLevel should be nenagtive integer \n');
    end
else
    opts.PrintLevel = 1;
end

% maximun iterations
if isfield(opts,'MaxIter')
    if ( opts.MaxIter <= 0 )
        error('opts.MaxIter should be positive integer \n');
    end
    if ( opts.MaxIter ~= fix(opts.MaxIter) )
        error('opts.PrintLevel should be integer \n');
    end
else
    opts.MaxIter = 200000;
end

% maximun function evaluations 
if isfield(opts,'MaxFunc')
    if ( opts.MaxFunc <= 0 )
        error('opts.MaxFunc should be integer \n');
    end
    if ( opts.MaxFunc ~= fix(opts.MaxFunc) )
        error('opts.MaxFunc should be integer \n');
    end
else
    opts.MaxFunc = 999999;
end

% tolerance for gradient norm
if isfield(opts,'tol')
    if ( opts.tol < 0 )
        error('opts.tol should be in positive \n');
    end
else
    opts.tol = 1e-6;
end

% lower bound for stepsizes 
if isfield(opts,'bbmin')
    if ( opts.bbmin <= 0 )
        error('opts.bbmin should be positive \n');
    end
else
    opts.bbmin = 1e-10;
end

% upper bound for stepsizes 
if isfield(opts,'bbmax')
    if ( opts.bbmax <= 0 )
        error('opts.bbmax should be positive \n');
    end
    if ( opts.bbmax <= opts.bbmin )
        error('opts.bbmax should be larger than opts.bbmin \n');
    end
else
    opts.bbmax = 1e6;
end

% adaptive parameter
if isfield(opts, 'tau') 
    if ( opts.tau <= 0 )
        error('opts.tau should be positive');
    end
else
    opts.tau = 0.2;     
end

% for updating adaptive parameter
if isfield(opts, 'gama')  
    if ( opts.gama < 1 )
        error('opts.gama should be larger or equal to 1');
    end
else
    opts.gama = 1.02;     
end

% parameter for recording history information
if isfield(opts,'record')
    if ( opts.record < 0 )
        error('opts.MaxFunc should be integer 0 or 1 \n');
    end
    if ( opts.record ~= fix(opts.record) )
        error('opts.MaxFunc should be integer 0 or 1 \n');
    end
else
    opts.record = 0;
end

% copy parameters
PrintLevel = opts.PrintLevel;
MaxIter = opts.MaxIter;
MaxFunc = opts.MaxFunc;
tol = opts.tol;
bbmin = opts.bbmin;
bbmax = opts.bbmax;
tau = opts.tau;
gama = opts.gama;
record = opts.record;

%% -------------------------------------------------------------------------
% initialize parameters for line search
if isfield(LinesearchParms,'delta')
    if ( LinesearchParms.delta <= 0 || LinesearchParms.delta >= 1 )
        error('delta should be in (0,1)\n');
    end
else
    LinesearchParms.delta = 0.0001;
end

if isfield(LinesearchParms,'sigma')
    if ( LinesearchParms.sigma <= 0 || LinesearchParms.sigma >= 1 )
        error('sigma should be in (0,1)\n');
    end
else
    LinesearchParms.sigma = 0.5;
end

if isfield(LinesearchParms,'M')
    if ( LinesearchParms.M <= 0 )
        error('LinesearchParms.M should be positive integer \n');
    end
    if ( LinesearchParms.M ~= fix(LinesearchParms.M) )
        error('LinesearchParms.M should be integer \n');
    end
else
    LinesearchParms.M = 10;
end

if isfield(LinesearchParms,'maxlsf')
    if ( LinesearchParms.maxlsf <= 0 )
        error('maxlsf should be positive integer \n');
    end
    if ( LinesearchParms.maxlsf ~= fix(LinesearchParms.maxlsf) )
        error('LinesearchParms.maxlsf should be integer \n');
    end
else
    LinesearchParms.maxlsf = 100;
end

M = LinesearchParms.M;
delta = LinesearchParms.delta;
sigma = LinesearchParms.sigma;
maxlsf = LinesearchParms.maxlsf;

%% -------------------------------------------------------------------------
iter = 0;
nfunc = 0;
ngrad = 0;

out.code = -1; 
out.iter = iter;
out.nfunc = nfunc;
out.ngrad = ngrad;

tic

x = x0;
x = max(LB,min(x,UB));
xbest = x;

fk = fobj(x); nfunc = nfunc + 1;
g = grad(x);  ngrad = ngrad + 1;
fbest = fk;

pg = min( max(LB, x - g), UB );
pgnorm = max(abs(pg - x));
pgnorm_best = pgnorm;

out.x = xbest;
out.obj = fbest;
out.pgnorm = pgnorm_best;
out.cputime = 0;
hist = [];
 
lambda = 1;

% to record recent M objective values
LastFvalues = zeros(M,1);
LastFvalues(1) = fk;
LastFvalues(2:end) = -1.0e99;

if ( PrintLevel >= 1 )
    logHead = ' %6s  %10s  %14s  %10s\n';
    logBody = ' %6i  %13.4e  %13.4e     %6.4f\n';
    fprintf(1,logHead,' iter','obj','pgnorm','step');
    if ( PrintLevel == 2 )
        fprintf(1,logBody,iter,fk,pgnorm,lambda);
    end
end

%% -------------------------------------------------------------------------
% test optimality condition for the initial point
CheckStop = 0;
if ( pgnorm <= tol )
    CheckStop = 1;
end
if ( CheckStop == 1 )
    if ( PrintLevel == 2 )
        fprintf(1,logBody,iter,fk,pgnorm,lambda);
    end
    out.code = 0;
    out.iter = iter;
    out.x = xbest;
    out.obj = fbest;
    out.pgnorm = pgnorm_best;
    out.cputime = toc;
    return
else
    xnorm = max(abs(x));
    if ( xnorm > 0.0 )
        alp = xnorm/pgnorm;
    else
        alp = 1.0/pgnorm;
    end
end

flag = 0;
mem_sty = zeros(2,1);
%% -------------------------------------------------------------------------
% main loop
while ( iter < MaxIter )
        
    % record history information
    if ( record == 1 )
        hist.bbs(iter) = alp;
        hist.step(iter) = lambda;
        hist.obj(iter) = fk;
        hist.gnorm(iter) = gnorm;
    end
    
	% test optimality condition 
    if ( pgnorm <= tol )
        CheckStop = 1;
    end
    if ( CheckStop == 1 )
        fbest = fk;
        xbest = x;
        pgnorm_best = pgnorm;
        out.code = 0;
        break
    end
    
    iter = iter + 1;
    
    if ( nfunc >= MaxFunc )
        out.code = 1;
        break
    end

    % GLL line search 
    px = min( max(LB, x - alp.*g), UB );
    d = px - x;
    gtd = g'*d;
    fr = max(LastFvalues);
    
    alpha = 1.0;
    nfunco = nfunc;
    while(1==1)
        xtemp = x + alpha.*d;
        ftemp = fobj(xtemp);
        nfunc = nfunc + 1;
        
        if ( ftemp <= (fr + delta*alpha*gtd) )
            break
        end
        
        if ( alpha <= 0.1)
            alpha = sigma * alpha;
        else
            atemp= (-gtd*alpha*alpha)/(2.0*(ftemp-fk-alpha*gtd));
            
            if ( atemp < 0.1 || atemp > (0.9*alpha))
                atemp = sigma * alpha;
            end
            
            alpha = atemp;
        end
        
        if ( nfunc >= (nfunco + maxlsf) )
            flag = 1;
            break
        end
    end
    
    lambda = alpha;
    if ( flag > 0 )
        out.code = 3;
        if ( PrintLevel > 1 )
            fprintf(1,'Line search failed.\n');
        end
        break
    end
   
    sx = xtemp - x;
    gold = g;
    
    % update x and g
    x = xtemp;
    fk = ftemp; 
    g = grad(x);
    ngrad = ngrad + 1;
    pg = min( max(LB, x-g), UB );
    pgnorm = max(abs(pg-x));  
    
    % record recent M objective values
    ii = mod(iter,M)+1;
    LastFvalues(ii) = fk;

    if ( PrintLevel == 2 )
        fprintf(1,logBody,iter,fk,pgnorm,lambda);
    end

    % update the best point and objective value
    if ( fk <= fbest )
        fbest = fk;
        xbest = x;
        pgnorm_best = pgnorm;
    end
    
    % compute new stepsize
    y = g - gold;    
    sty = sx'*y;
    if ( sty > 0 )
        mem_sty(mod(iter,2)+1) = 1;
        if ( iter>0 && mem_sty(mod(iter-1,2)+1)==1 )
            bb1_old = bb1;
            bb2_old = bb2;
        end
        
        y2 = y(logical(sx));
        bb1 = (sx'*sx)/sty;
        bb2 = (sty)/(y2'*y2);     
        if ( iter > 1 )
            if ( bb2/bb1 < tau && sum(mem_sty)==2 )
                phi13 = (bb2_old - bb2)/(bb2_old * bb2 * (bb1_old - bb1));
                phi23 = phi13 * bb1 + 1/bb2;
                bb_new = 2/( phi23 + sqrt(phi23^2 - 4*phi13) );
                if ( bb_new <= 0 )
                    alp = min(bb2_old,bb2);
                else
                    alp = min([bb_new,bb2,bb2_old]);
                end
                tau = tau/gama;
            else
                alp = bb1;
                tau = tau*gama;
            end
        else
            alp = bb1;
        end
    else
        mem_sty = zeros(2,1);
        xnorm = max(abs(x));
        alp = min(1.0/pgnorm, xnorm/pgnorm);
    end
    alp = max(bbmin,min(alp,bbmax));     
end

if ( iter == MaxIter )
    out.code = 2;
    if ( PrintLevel > 1 )
        fprintf(1,'Maximum number of function evaluations reached.\n');
    end
end

if ( PrintLevel >= 1 )
    fprintf(1,logBody,iter,fk,pgnorm,lambda);
end

hist.bbs(iter) = alp;
hist.step(iter) = lambda;
hist.obj(iter) = fk;
hist.pgnorm(iter) = pgnorm;

out.iter = iter;
out.nfunc = nfunc;
out.ngrad = ngrad;
out.x = xbest;
out.obj = fbest;
out.pgnorm = pgnorm_best;
out.cputime = toc;